
%option noyywrap
%option align fast

/* Whitespace, excluding newline. */
W	[ \t]

/* Non-whitespace. */
NW	[^ \t\n]

/* Non-whitespace, non-equal. */
NWE	[^ \t\n=]

/* Non-whitespace, non-colon. */
NWC	[^ \t\n:]

/* Uppercase only */
UC	[A-Z]

/* Uppercase and digits */
UCD	[A-Z0-9]

/* Any letters or digits */
LD	[a-zA-Z0-9]

/* Start conditions: */
%x wantSendmail1 wantSendmail2 wantSendmail3 wantQID endQID
%x inKey inValue inQuote inBracket
%x inClone inStat inStatMaybeQid
%x DiscardLine UnclasLine RestOfValue TextIsNote
%x ValueEscape QuoteEscape BracketEscape

%{
    #define QID_LEN	(40)	/* Crucial not to undersize this!
    				 * Must be long enough to hold QID plus
				 * trailing null.
				 */

    #include <stdio.h>
    #include <errno.h>
    #include <stdlib.h>
    #include <string.h>
    #include <unistd.h>
    #include <sys/stat.h>
    #include <sys/time.h>
    #include <search.h>
    #include <unistd.h>

    static char *Prog = "";
    static int Debug = 0;
    static int MakeRelationShips = 0;
    static int Nqid = 0;	/* Number unique qid's seen */
    static struct QIDHash **QIDtbl;	/* table of qid item ptrs */
    static int QIDtblsize=0;		/* length of QIDtbl */

    /* A struct suitable for saving (yytext, yyleng). */
    struct Saver {
	char *text;
	int leng;	/* length of useful data in text */
	int buflen;	/* length of text buffer; >= leng */
    };
    typedef struct Saver Saver;

    /* A struct suitable for saving a qid, and a pointer to another struct */
    struct QIDElt {
	char qid[QID_LEN];
	struct QIDElt *next;
    };
    typedef struct QIDElt QIDElt;

    /* The struct that we put on the hash list */
    struct QIDHash {
	int filenum;		/* for tracking last-file, last-line number */
	int linenum;		/* for tracking last-file, last-line number */
	char qid[QID_LEN];	/* this qid */
	char clone_of[QID_LEN];	/* this elt is clone of clone_of. */
	QIDElt *child;		/* this elt has these children */
    };
    typedef struct QIDHash QIDHash;

    /* Where our functions can record trouble messages */
    static char *errorMsg = NULL;

    /* Where we save the start-of-line info (before we have an indexing
     * QID available).
     */
    static Saver LinePreamble = {NULL, 0, 0};	/* mon day time host etc */
    static Saver LinePid = {NULL, 0, 0};	/* sendmail pid */
    static Saver LineQID = {NULL, 0, 0};	/* QID -- id in mail queue */

    static Saver Key = {NULL, 0, 0};		/* For holding a key */
    static Saver Value = {NULL, 0, 0};		/* For building up a value */
    static Saver Unclas = {NULL, 0, 0};		/* For unclassified text */

    /* Functions to place text into a Saver struct */
    static int saver(Saver *sp, int append, char *mytext, int myleng);

    /* Convenience definition for saving yytext */
    #define save_yytext(sp, append) saver(sp, append, yytext, yyleng)

    /* Function to initialize output for a line */
    static int init_line(char *QID);

    /* Function to record a qid's last record value */
    static int last_line(char *QID);

    /* Function to record child info */
    static void save_child_info(char *QID, char *child_qid, int child_qid_l);

    /* Function to record clone info */
    static void save_clone_info(char *QID, char *clone_qid, int clone_qid_l);

    /* Function to store key/value */
    static int emit_keyvalue(void);

    static int FileNum=1;
    static int LineNum=0;

    static int collectKeyVal=0;		/* Collect key/value? */

    static int collectLastLine=0;	/* Collect last-line info? */

    static int doEmit=1;		/* controls emit_keyvalue() */
%}

%%

<DiscardLine>.*\n	{
    /* Eat up line, throw away, reset for new line. */
    BEGIN(INITIAL) ;
}

<RestOfValue>.*\n	{
    /* Append rest of line to value; save; reset for new line. */
    if (save_yytext(&Value, 1)) yyterminate();
    Value.leng--;	/* Discard the trailing newline */
    if (emit_keyvalue()) yyterminate();
    BEGIN(INITIAL) ;
}

<UnclasLine,TextIsNote>.*\n	{
    /* In either case, we will just force the entire value string to
     * have a keyword "notes".  Write out the value and reset for a newline.
     * (Ignore line if on first pass.)
     */
    if (collectKeyVal) {
	if (saver(&Key, 0, "notes", -1)) yyterminate();
	if (save_yytext(&Value, 1)) yyterminate();
	Value.leng--;	/* Discard the trailing newline */
	if (emit_keyvalue()) yyterminate();
    }
    BEGIN(INITIAL) ;
}

^{NW}+{W}+{NW}+{W}+{NW}+{W}+{NW}+{W}+sendmail {
			    /* At beginning of line, match:
			     *		mon day time host sendmail
			     * Then set start cond. to match `[pid]:? QID:'
			     * or match `[pid]:? [ID xxx facility.level] QID:'
			     */
			    ++LineNum;
			    LinePid.leng = 0;	/* mark Pid unknown */
			    LineQID.leng = 0;	/* mark QID unknown */
			    Key.leng = 0;	/* mark key unknown */
			    Value.leng = 0;	/* mark value unknown */
			    if (save_yytext(&LinePreamble, 0)) yyterminate();
			    BEGIN(wantSendmail1);
			}

^.			{ /* Not a sendmail-type line. */
			    ++LineNum;
			    if (Debug > 1) {
				fprintf(stderr, "DBG: discard %d\n", LineNum);
			    }
			    BEGIN(DiscardLine);
			}
^\n			{
			    /* Not a sendmail-type line */
			    ++LineNum;
			}

<wantSendmail1>"["	{
			    /* expected text */
			    BEGIN(wantSendmail2);
			    if (Debug > 1) {
				fprintf(stderr, "DBG match 1 %d\n", LineNum);
			    }
			}
<wantSendmail1>[^\n]	{ yymore(); BEGIN(UnclasLine); }

<wantSendmail2>[0-9]+	{
			    /* Matched PID section of line */
			    if (Debug > 1) {
				fprintf(stderr, "DBG match 2 %d\n", LineNum);
			    }
			    if (save_yytext(&LinePid, 0)) yyterminate();
			    BEGIN(wantSendmail3);
			}
<wantSendmail2>[^\n]	{ yymore(); BEGIN(UnclasLine); }

<wantSendmail3>"]":?{W}*("["[^]]*"]")?{W}+	{
			    BEGIN(wantQID);
			    if (Debug > 1) {
				fprintf(stderr, "DBG match 3 %d\n", LineNum);
			    }
			}
<wantSendmail3>[^\n]	{ yymore(); BEGIN(UnclasLine); }

<wantQID>{LD}+		{ 
			    /* Matched QID section of line */
			    if (Debug > 1) {
				fprintf(stderr, "DBG match QID %d\n", LineNum);
			    }
			    if (save_yytext(&LineQID, 0)) yyterminate();
			    BEGIN(endQID);
			}
<wantQID>[^\n]		{ yymore(); BEGIN(UnclasLine); }
<endQID>:{W}+		{
			    /* Matched post-QID section of line */
			    if (Debug > 1) {
				fprintf(stderr, "DBG match post-QID %d\n", LineNum);
			    }
			    if (collectLastLine) {
				if (last_line(LineQID.text)) yyterminate();
			    }
			    if (collectKeyVal) {
				/* Collect keys and values */
				if (init_line(LineQID.text) < 0) yyterminate();
				BEGIN(inKey);
			    } else {
				BEGIN(DiscardLine);
			    }
			}
<endQID>[^\n]		{ yymore(); BEGIN(UnclasLine); }

<inKey>stat=		{
			    /* Identified the "stat" key, which takes arbitrary
			     * text up to newline.
			     */
			    if (save_yytext(&Key, 0)) yyterminate();
			    Value.leng = 0;
			    BEGIN(inStat);
			}
<inStat>"Sent ("	{
			    /* This might match stat=Sent (EAA12345 ...)
			     * (We don't use trailing context to check for the
			     * qid, in an effort to keep peak performance.)
			     */
			    if (save_yytext(&Value, 1)) yyterminate();
			    BEGIN(inStatMaybeQid);
			}
<inStat>.		{
			    /* Rest of text is simple stat line. */
			    if (save_yytext(&Value, 1)) yyterminate();
			    BEGIN(RestOfValue) ;
			}
<inStatMaybeQid>[A-Z]+[0-9]+	{
			    /* We got stat=Sent (qid ...).  Output the
			     * child-is/qid, and then treat the line as plain
			     * stat text.
			     */
			    if (doEmit) {
				printf("child-is\n%.*s\n", yyleng, yytext);
			    } else {
				/* If not emitting value, we must be
				 * collecting "heritage" (parent/child) info.
				 */
				save_child_info(LineQID.text, yytext, yyleng);
			    }
			    if (save_yytext(&Value, 1)) yyterminate();
			    BEGIN(RestOfValue);
			}
<inStatMaybeQid>.	{
			    /* It didn't contain a qid; it's a simple stat */
			    if (save_yytext(&Value, 1)) yyterminate();
			    BEGIN(RestOfValue) ;
			}
<inKey>{NWE}+=		{
			    /* Identified the key; begin a value */
			    if (save_yytext(&Key, 0)) yyterminate();
			    Value.leng = 0;
			    BEGIN(inValue);
			}
<inKey>"clone "		{
			    /* Identified a "clone" note, like
			     *     clone HAA14081, owner=request-nics
			     * Start it as a "notes" item, but also
			     * set state to inClone.
			     */
			    if (saver(&Key, 0, "notes", -1)) yyterminate();
			    if (saver(&Value, 0, "clone ", -1)) yyterminate();
			    BEGIN(inClone);
			}
<inClone>{UCD}+		{
			    /* We got "clone QID".  Output the
			     * clone-of/qid, and then treat the line as plain
			     * clone text.
			     */
			    if (doEmit) {
				printf("clone-of\n%.*s\n", yyleng, yytext);
			    } else {
				/* If not emitting value, we must be
				 * collecting "heritage" (parent/child) info.
				 */
				save_clone_info(LineQID.text, yytext, yyleng);
			    }
			    if (save_yytext(&Value, 1)) yyterminate();
			    BEGIN(RestOfValue);
			}
<inClone>.		{
			    /* It didn't contain a qid.  Just output it. */
			    if (save_yytext(&Value, 1)) yyterminate();
			    BEGIN(RestOfValue) ;
			}

<inKey>{NWE}+[ \t]	{
			    /* Whoops -- In the place we expected to see
			     * a key=value, we didn't see that equal sign.
			     * Just append the whole thing to "notes".
			     */
			    yymore(); BEGIN(TextIsNote);
			}
    
<inValue>[^\n,\"<\\]+	{
			    if (save_yytext(&Value, 1)) yyterminate();
			}
<inValue>\n		{
			    if (emit_keyvalue()) yyterminate();
			    BEGIN(INITIAL);
			}
<inValue>", "		{
			    if (emit_keyvalue()) yyterminate();
			    BEGIN(inKey);
			}
<inValue>,		{
			    if (save_yytext(&Value, 1)) yyterminate();
			}
<inValue>\"		{
			    if (save_yytext(&Value, 1)) yyterminate();
			    BEGIN(inQuote);
			}
<inValue>"<"		{
			    if (save_yytext(&Value, 1)) yyterminate();
			    BEGIN(inBracket);
			}

<inValue,inQuote,inBracket>\\\n	{
			    /* Line continuation -- discard
			     * the backslash-newline.
			     */
			}
<inValue>\\		{ BEGIN(ValueEscape); /* escape next character */ }
<inQuote>\\		{ BEGIN(QuoteEscape); /* escape next character */ }
<inBracket>\\		{ BEGIN(BracketEscape); /* escape next character */ }

<ValueEscape>.		{
			    /* character after a backslash */
			    if (save_yytext(&Value, 1)) yyterminate() ;
			    BEGIN(inValue);
			}

<QuoteEscape>.		{
			    /* character after a backslash */
			    if (save_yytext(&Value, 1)) yyterminate() ;
			    BEGIN(inQuote);
			}

<BracketEscape>.	{
			    /* character after a backslash */
			    if (save_yytext(&Value, 1)) yyterminate() ;
			    BEGIN(inBracket);
			}

<inQuote>[^\"\\\n]+	{
			    if (save_yytext(&Value, 1)) yyterminate() ;
			}
<inQuote>[\"]		{
			    if (save_yytext(&Value, 1)) yyterminate() ;
			    BEGIN(inValue);
			}

<inBracket>[^\>\\\n]+	{
			    if (save_yytext(&Value, 1)) yyterminate() ;
			}
<inBracket>[\>]		{
			    if (save_yytext(&Value, 1)) yyterminate() ;
			    BEGIN(inValue);
			}

<*>\n			{
			    /* Shouldn't get a newline here...
			     * Save whatever we've got as a value,
			     * then start next line.
			     */
			    if (emit_keyvalue()) yyterminate();
			    BEGIN(INITIAL) ;
			}
<*>.			{ yymore() ; BEGIN(UnclasLine); }

%%

void
usage() {
    fprintf(stderr, "Use: %s [-V] [-d] file...\n", Prog);
}

double difftimeval(struct timeval *t1, struct timeval *t2);

int
main(int argc, char **argv)
{
    char *file;
    char *s, *opt;
    int pass, i, iarg, firstfilearg, save_LineNum, par_child, showProgress;
    struct stat statfile;
    long totsize, n;
    struct timeval tv0, tv1;
    double t;
    int doArgs(int argc, char **argv, int *debug, int *sp, int *parent_child);
    int loopOverFiles(char **Files, int nfiles, int sp);
    void outputParentChild();

    Prog = (s = strrchr(argv[0], '/')) ? s+1 : argv[0];

    firstfilearg = doArgs(argc, argv, &Debug, &showProgress, &par_child);
    if (Debug) {
	fprintf(stderr,
	    "%s: $Id: maillex.l,v 1.43 2002/03/06 05:40:25 will Exp $\n",
	    Prog);
    }

    if (firstfilearg == -1)
	return 1;


    /* First, determine size of input files, and use this to estimate
     * the number of lines (just assume there are 100 chars/line, which
     * is conservative for sendmail logs).
     */
    for (totsize = 0, iarg = firstfilearg; iarg < argc; iarg++) {
	file = argv[iarg];
	if (stat(file, &statfile) == -1) {
	    fprintf(stderr, "%s: can't stat \"%s\": ", Prog, file);
	    perror("");
	    return(1);
	}
	totsize += statfile.st_size;
    }
    n = totsize/100;
    if (hcreate(n) == 0) {
	fprintf(stderr, "%s: can't create hash table of size %d\n", Prog, n);
	return (1);
    }
    /* Malloc space for qid pointers */
    QIDtbl = (QIDHash **) malloc(n*sizeof(*QIDtbl));
    if (!QIDtbl) {
	fprintf(stderr, "%s: can't create qid table of size %d\n", Prog, n);
	return (1);
    }
    QIDtblsize = n;

    /* Loop over filename arguments, processing each one.
     * We use two passes: on the first one, we identify all
     * the last entries for each id; on the second pass, we
     * analyze the entire text and send the analyzed form out.
     *
     * Timing shows that the first pass is about 6 times quicker than
     * the second pass, which is why we are careful not to waste time
     * parsing the file on the first pass.
     */

    /* Make two passes through each file: once to identify the last instance
     * of each QID (and collect parent/child info, if desired), and once
     * to collect keys and values.
     */

    /* PASS 1 */
    gettimeofday(&tv0, NULL);

    collectLastLine = 1;	/* On first pass, get last-line info. */
    collectKeyVal = par_child;	/* collect keyval iff want parent/child info */
    doEmit = 0;			/* don't emit values */

    if (!loopOverFiles(&argv[firstfilearg], argc-firstfilearg, showProgress))
	return 1;

    gettimeofday(&tv1, NULL);
    t = difftimeval(&tv0, &tv1);
    if (Debug)
	fprintf(stderr, "Pass 1: %.3f sec\n", t);

    /* ----------------------------------------------------------------- */
    if (par_child) {
	outputParentChild();
    }
    /* ----------------------------------------------------------------- */

    /* PASS 2 */
    gettimeofday(&tv0, NULL);

    collectLastLine = 0;	/* Pass 1 got last-line info */
    collectKeyVal = 1;		/* definitely need these! */
    doEmit = 1;			/* and it's time to print */

    if (!loopOverFiles(&argv[firstfilearg], argc-firstfilearg, 0))
	return 1;

    /* Finish up by printing the last line number.  This gives the caller
     * an accurate line count, even if the last line of text isn't the last
     * sendmail line.
     */
    printf("line\n%d\n", LineNum);

    gettimeofday(&tv1, NULL);
    t = difftimeval(&tv0, &tv1);
    if (Debug)
	fprintf(stderr, "Pass 2: %.3f sec\n", t);

    return 0;
}


/* *********************************************************************** */
/*
 * Process options, Return firstfilearg index on success, -1 on problem.
 */
int
doArgs(int argc, char **argv, int *debug, int *showProgress, int *parent_child)
{
    int iarg;
    char *opt;

    *debug = 0;		/* default values */
    *parent_child = 0;
    *showProgress = 0;

    for (iarg=1; argv[iarg] && argv[iarg][0] == '-'; iarg++) {
	for (opt = &argv[iarg][1]; *opt; opt++) {
	    switch (*opt) {
	    case 'd':
		*debug += 1;
		break;
	    case 'p':
		*parent_child = 1;
		break;
	    case 'P':
		*showProgress = 1;
		break;
	    case 'V':
		/* show version and exit */
		fprintf(stderr, "%s: $Id: maillex.l,v 1.43 2002/03/06 05:40:25 will Exp $\n", Prog);
		exit(0);
		break;
	    default:
		fprintf(stderr, "Unknown option `%c'\n", *opt);
		usage();
		return -1;
	    }
	}
    }
    if (iarg >= argc) {
	usage();
	return -1;
    }
    return iarg;
}

/* *********************************************************************** */
/*
 * Apply the lexer to a set of files.
 * Return 1 on success, 0 if there is trouble.
 */
int
loopOverFiles(
    char **Files,
    int nfiles,
    int showProgress)
{
    char *file;

    for (FileNum=0; FileNum < nfiles; FileNum++) {
	LineNum = 0;
	file = Files[FileNum];

	yyin = fopen(file, "r");
	if (!yyin) {
	    fprintf(stderr, "%s: can't open \"%s\": ", Prog, file);
	    perror("");
	    return 0;
	}
	if (showProgress) {
	    fprintf(stderr, "\rScanning file %d... ", FileNum+1);
	    fflush(stderr);
	}
	yylex();
	fclose(yyin);
	if (showProgress) {
	    fprintf(stderr, "\n");
	}
	if (errorMsg) {
	    fprintf(stderr, "%s: %s\n", Prog, file);
	    return 0;
	}
    }
    return 1;
}

/* *********************************************************************** */
/*
 * Return difference, in sec, of two timeval structs.
 * Returned value is t2 - t1.
 */
double
difftimeval(struct timeval *t1, struct timeval *t2)
{
    return (double) (t2->tv_sec - t1->tv_sec) +
    		(double) (t2->tv_usec - t1->tv_usec) / 1.0e6;
}

/* *********************************************************************** */
/*
 * Send the recorded child-is and clone-of structure to stdout.
 */
void
outputParentChild()
{
    int i;
    QIDElt *qp;

    /* clone info */
    printf("array set CloneInfo {\n");
    for (i=0; i < Nqid; i++) {
	if (QIDtbl[i]->clone_of[0])
	    printf("%s %s\n", QIDtbl[i]->qid, QIDtbl[i]->clone_of);
    }
    printf("}\n\n");

    /* child info */
    printf("array set ChildInfo {\n");
    for (i=0; i < Nqid; i++) {
	if (QIDtbl[i]->child) {
	    printf("%s { ", QIDtbl[i]->qid);
	    for (qp = QIDtbl[i]->child; qp; qp = qp->next)
		printf("%s ", qp->qid);
	    printf("}\n");
	}
    }
    printf("}\n");
}

/* *********************************************************************** */
/*
 * This function should be called (during the second pass) after everything
 * through the QID has been read: we have something that is sufficiently like
 * a proper sendmail line that we want to record all of it.
 *
 * Return 0 on success, 1 on ignore-this-line, -1 otherwise.
 */
static int
init_line(char *QID)
{
    static Saver mon={NULL, 0, 0};
    static Saver day={NULL, 0, 0};
    static Saver time={NULL, 0, 0};
    static Saver host={NULL, 0, 0};
    int linenum, filenum;
    ENTRY *ep, e;

    /* Find the lastline entry.  If it's missing, it must mean that the
     * line was added in the time gap between the first and second
     * passes; ignore the line.
     */
    e.key = QID;
    if (!(ep = hsearch(e, FIND))) {
	return 1;
    }
    filenum = ((QIDHash *) ep->data)->filenum;
    linenum = ((QIDHash *) ep->data)->linenum;

    /* Ensure that all the Savers are big enough.
     * The easy way to do this is to append a 0-length string
     * onto each one.
     */
    if (saver(&mon,  1, "", 0)) return -1;
    if (saver(&day,  1, "", 0)) return -1;
    if (saver(&time, 1, "", 0)) return -1;
    if (saver(&host, 1, "", 0)) return -1;

    if (sscanf(LinePreamble.text, "%s %s %s %s", mon.text,
			day.text, time.text, host.text) != 4) {
	/* can't happen :-) */
	errorMsg = "Internal line-parsing inconsistency!";
	return -1;
    }
    mon.leng = strlen(mon.text);
    day.leng = strlen(day.text);
    time.leng = strlen(time.text);
    host.leng = strlen(host.text);

    if (doEmit) {
	/* "line" must come first: it's used by maillog to mark the boundary
	 * between entries for line N and N+1.
	 */
	printf("line\n%d\n", LineNum);
	printf("lastline\n%d\n", (LineNum == linenum && FileNum == filenum));

	/* qid must come before any keys besides "line" and "lastline",
	 * because maillog uses the qid when storing any of those keys.
	 */
	printf("qid\n%s\n", LineQID.text);
	printf("filenum\n%d\n", FileNum+1);	/* use FileNum+1 so it's 1..N */
	printf("date\n%s %s %s\n", mon.text, day.text, time.text);
	printf("host\n%s\n", host.text);
	printf("pid\n%s\n", LinePid.text);
    }
    
    return 0;
}

/* *********************************************************************** */
/*
 * This function should be called (during the first pass) for each line read
 * in, after everything through the QID has been read: we have something that
 * is sufficiently like a proper sendmail line that we want to identify this
 * as the last known line for this QID.
 *
 * Return 0 on success, -1 otherwise.
 */
static int
last_line(char *QID)
{
    ENTRY e, *ep;

    e.key = QID;
    if ((ep = hsearch(e, FIND))) {
	/* Found an existing entry; update it */
	((QIDHash *) ep->data)->filenum = FileNum;
	((QIDHash *) ep->data)->linenum = LineNum;
    } else {
	/* Need to create an entry.  For malloc efficiency, allocate
	 * a single block that is big enough for the QID, line number,
	 * clone, and child info. Put the linenum first so as to assure
	 * its alignment.
	 */
	char *s;
	QIDHash *p;
	if (!(s = malloc(sizeof(QIDHash)))) {
	    fprintf(stderr, "Failed to malloc hash table entry at "
	    		"filenum %d, record %d\n", FileNum, LineNum);
	    return -1;
	}
	p = (QIDHash *) s;
	p->filenum = FileNum;
	p->linenum = LineNum;
	strcpy(p->qid, QID);
	p->clone_of[0] = '\0';
	p->child = NULL;

	e.data = s;
	e.key = p->qid;		/* note that the key is stored as a field
				 * in the data.
				 */

	QIDtbl[Nqid] = p;	/* save location of this qid */
	Nqid++;
	if (Nqid == QIDtblsize) {
	    QIDtbl = realloc(QIDtbl, 2*Nqid*sizeof(*QIDtbl));
	    if (!QIDtbl) {
		fprintf(stderr, "%s: can't create qid table of size %d\n",
					Prog, Nqid*2);
		exit(1);
	    }
	    QIDtblsize *= 2;
	}

	if (!hsearch(e, ENTER)) {
	    fprintf(stderr, "Failed to insert last-line entry at record %d\n",
	    	LineNum);
	}
    }
    
    return 0;
}

/* *********************************************************************** */
/*
 * This function should be called (during the first pass) when a clone-of
 * field is encountered.  We update the hash table entry with these values.
 *
 * Return 0 on success, -1 otherwise.
 */
static void
save_clone_info(
    char *QID,
    char *clone_of,
    int clone_of_l)
{
    ENTRY e, *ep;

    e.key = QID;
    if (!(ep = hsearch(e, FIND))) {
	/* Didn't find this entry ?! */
	fprintf(stderr, "save_clone_info() found no entry for qid=%s?!\n",
	    	QID);
	exit(1);
    }

    strncpy(((QIDHash *) ep->data)->clone_of, clone_of, clone_of_l);
    ((QIDHash *) ep->data)->clone_of[clone_of_l] = '\0';
}

/* *********************************************************************** */
/*
 * This function should be called (during the first pass) when a child-is
 * field is encountered.  We update the hash table entry with these values.
 *
 * Return 0 on success, -1 otherwise.
 */
static void
save_child_info(
    char *QID,		/* qid to find */
    char *child_qid,	/* NULL means no child info */
    int child_qid_l)
{
    ENTRY e, *ep;
    QIDElt *newchild;
    QIDHash *item;

    e.key = QID;
    if (!(ep = hsearch(e, FIND))) {
	/* Didn't find this entry ?! */
	fprintf(stderr, "save_child_info() found no entry for qid=%s?!\n",
	    	QID);
	exit(1);
    }
    item = (QIDHash *) ep->data;

    /* Add new child info */
    newchild = (QIDElt *) malloc(sizeof(QIDElt));
    if (!newchild) {
	fprintf(stderr, "malloc failed for QIDElt!\n");
	exit(1);
    }
    strncpy(newchild->qid, child_qid, child_qid_l);
    newchild->qid[child_qid_l] = '\0';
    newchild->next = item->child;
    item->child = newchild;
}

/* *********************************************************************** */
/* Emit a key + value.
 *
 * Return 0 on success, -1 otherwise.
 */
static int
emit_keyvalue() {
    char *s;
    if (Key.leng && Key.text[Key.leng-1] == '=')
	Key.leng--;
    if ((s = strchr(Value.text, '\n')))
	*s = '\n';	/* just in case */
    if (doEmit) {
	printf("%.*s\n%.*s\n", Key.leng, Key.text, Value.leng, Value.text);
    }
    return 0;
}

/* *********************************************************************** */
/*
 * Save (or append) some text.
 * We always add null-terminator, but the function is 8-bit clean.
 * Normally, myleng characters are copied.  However, to make handling
 * null-terminated strings convenient, we also use the rule that if
 * myleng < 0, we use strlen(mytext).
 *
 * Returns 0 on success, -1 on trouble.
 */

static int
saver(Saver *sp, int append, char *mytext, int myleng)
{
    int l;
    char *p;

    if (myleng < 0)
	myleng = strlen(mytext);

    if (!append)
	sp->leng = 0;

    if (!sp->text || sp->buflen == 0) {
	/* Saver doesn't yet have any buffer at all */
	l = 2*myleng+1;
	sp->text = malloc(l);
	if (!sp->text) {
	    errorMsg = "malloc failure";
	    return -1;
	}
	sp->leng = 0;
	sp->buflen = l;
    } else if (append && (myleng + sp->leng) >= sp->buflen) {
	/* Need room to append */
	l = 2*(myleng + sp->leng) + 1;
	p = realloc(sp->text, l);
	if (!p) {
	    errorMsg = "malloc failure";
	    return -1;
	}
	sp->text = p;
	sp->buflen = l;
    } else if (!append && myleng >= sp->buflen) {
	/* Need more room to overwrite */
	l = 2*myleng + 1;
	p = realloc(sp->text, l);
	if (!p) {
	    errorMsg = "malloc failure";
	    return -1;
	}
	sp->text = p;
	sp->leng = 0;
	sp->buflen = l;
    }
    if (myleng > 0) {
	memcpy(sp->text+sp->leng, mytext, myleng);
	sp->leng += myleng;
    }
    sp->text[sp->leng] = '\0';
#if 0
    fprintf(stderr, "saver(%p)=(%d) %.*s\n", sp, sp->leng, sp->leng, sp->text);
#endif
    return 0;
}

